#include <stdlib.h>
#include "sage++user.h"

#define IN_BLOB_CODE_C
#include "Blob_Code.h"

Blob_List::Blob_List (SgStatement *s, int first_or_second)
{
     which_blob_list = first_or_second;
     our_parent = s;
     
     if ( first_or_second == 1 )
     {
	  our_list = BIF_BLOB1(our_parent->thebif);
     }
     else if ( first_or_second == 2 )
     {
	  our_list = BIF_BLOB2(our_parent->thebif);
     }
     else
     {
	  /* Impossible condition */
	  abort();
     }
}

/*
  Start at element zero and walk along until we
  get to the requested element (i steps past 0).
  */

PTR_BLOB
Blob_List::elt(int i)
{
     PTR_BLOB current = our_list;

     for (int j = 1; j <= i ; j++)
     {
	  if ( current != 0 )
	       current = BLOB_NEXT(current);
     }

     return current;
}

PTR_BFND
Blob_List::value(int i)
{
     PTR_BLOB the_elt = elt(i);

     if ( the_elt != 0 )
	  return BLOB_VALUE(the_elt);
     else
	  return 0;
}

/*
  length() is the number of elements in the list.

  Members of the list are: 0 .. length()-1
  */

int
Blob_List::length(void)
{
     PTR_BLOB current = our_list;
     PTR_BLOB next;
     int count;

     if ( current == 0 )
	  return 0;

     count = 1;
     next = BLOB_NEXT(current);

     while (next != 0)
     {
	  current = next;
	  next = BLOB_NEXT(current);
	  count++;
     }

     return count;
}

void
Blob_List::insert_after (PTR_BLOB location, PTR_BLOB item)
{
     if ( location == 0 )
	  abort(); /* Bad location */

     PTR_BLOB next = BLOB_NEXT(location);

     BLOB_NEXT(location) = item;
     BLOB_NEXT(item) = next;
}
     

void
Blob_List::insert(PTR_BLOB item, int location)
{
     if ( location == 0 )
     {
	  BLOB_NEXT(item) = our_list;
	  our_list = item;

	  if ( which_blob_list == 1 )
	  {
	       BIF_BLOB1(our_parent->thebif) = our_list;
	  }
	  else /* ( which_blob_list == 2 ) */
	  {
	       BIF_BLOB2(our_parent->thebif) = our_list;
	  }
     }
     else /* (location != 0 ) */
     {
	  /* before is guaranteed to be valid, as long as
	     location <= this->length() */
	  PTR_BLOB before = elt(location - 1);

	  insert_after (before, item);
     }
}

int
Blob_List::index (SgStatement *the_stmt)
{
     Blob_List_Iterator iterator(this);
     int current_index = 0;

     for ( iterator.reset();
	   ! iterator.done();
	   iterator.advance())
     {
	  if (BLOB_VALUE(iterator.current()) == (the_stmt->thebif))
	       return current_index;

	  current_index++;
     }
     return -1;
}

int
Blob_List::contains (SgStatement *the_stmt)
{
     return (index(the_stmt) != -1 );
}

