// This may look like C code, but it is really -*- C++ -*-

#ifndef BLOB_CODE_H
#define BLOB_CODE_H 1

#include "sage++user.h"

/*

  Make a Blob List class to bring stuff up to the Sage++ level.
  
  Grabs the list of Blob nodes hanging off a given statement.

  We do not allow null statement pointers in the constructor.
  */


class Blob_List {
  public:
     Blob_List (SgStatement *s, int first_or_second);

     /* Access begins with i = 0 */
     PTR_BLOB elt(int i);
     PTR_BFND value (int i);

     /* Members in the list range from 0 ... length()-1 */
     int length(void);
     
     /* Location gives the location to insert before. To insert at
	the start of the list, use location = 0; to insert at the end,
	use location == length().
	*/
     void insert(PTR_BLOB item, int location);

     /* Inserts ITEM after LOCATION. The idea is that LOCATION will
	be a member of this list, but that is not strictly necessary. */
     
     void insert_after (PTR_BLOB location, PTR_BLOB item);

     /* Returns index of THE_STMT, or -1 if not found. */
     int index (SgStatement *the_stmt);

     /* Returns 1 if THE_STMT is in our list, 0 otherwise. */
     int contains (SgStatement *the_stmt);
     
  private:
     SgStatement *our_parent;
     int which_blob_list;
     PTR_BLOB our_list;
};



/***********************************************************************/

class Blob_List_Iterator {
  public:
     Blob_List_Iterator (Blob_List *the_list);
     
     /* Return current or next: does not move iteration. */
     PTR_BLOB current(void);
     PTR_BLOB next(void);

     /* Reset current to start */
     void reset(void);

     /* Move current to next */
     PTR_BLOB advance(void);

     int done(void);
  private:
     Blob_List *our_list;
     PTR_BLOB our_current;
};

#include "Blob_Code.icc"

#endif /* ndef BLOB_CODE_H */
