// This may look like C code, but it is really -*- C++ -*-

/*
  WARNING: None of the FORTRAN_LANG stuff is tested.
 */

#include "sage++user.h"
#include "iterator.h"

void
Iterator::begin (void)
{
     down_namescope(SOFT_AND_HARD);

     /* Fill it up with the global symbols in this file. */
     namescope->grab_all_symbols (the_file->firstStatement());

     /* Do stuff */
     iterate (the_file->firstStatement());

     up_namescope (SOFT_AND_HARD);
}


/*
  Namescope managing functions. These member functions somewhat resemble
  those found in variablename.C

  */

/* Note that going DOWN, we should build the hard namescope first,
   then the soft one, whereas they should be destroyed in the reverse
   order. */

void
Iterator::down_namescope (enum Scope_Movement which /* = SOFT_AND_HARD */)
{
     if ( which & HARD )
     {
	  hard_namescope = new Symbol_Maintainer (hard_namescope);
     }

     if ( which & SOFT )
     {
	  namescope = new Symbol_Maintainer (namescope, hard_namescope);
     }
}

void
Iterator::up_namescope (enum Scope_Movement which /* = SOFT_AND_HARD */)
{
     Symbol_Maintainer *up_one;

     if ( which & SOFT )
     {
	  if (namescope != NULL)
	  {
	       up_one = namescope->parent;
	       delete namescope;
	       
	       namescope = up_one;
	  }
     }

     if ( which & HARD )
     {
	  if (hard_namescope != NULL)
	  {
	       up_one = hard_namescope->parent;
	       delete hard_namescope;
	       
	       hard_namescope = up_one;
	  }
     }
}



/*
  Iterator::iterate

  One stupid thing that I want to mention: We keep track of the current
  and next statement in local variables so we can call ourselves recursively.
  However, other routines may want this information (for instance
  if up_namescope() is overridden to insert code at the end of each
  basic block), so at various points I copy the local information back
  to the class-"global" vars "current" and "next".

 */

#define UPDATE_CURRENT() current = current_stmt; next = next_stmt;

void
Iterator::iterate (SgStatement *current_stmt)
{
     SgStatement *next_stmt;
     
     while ( current_stmt != NULL )
     {
          /* We have to find the next statement first thing, because
             one of the actions we may want to perform is deleting 
             the current statment! Needless to say, this screws 
             up our ability to find the next statement later. */

          next_stmt=current_stmt->nextInChildList();

	  /* Update the current and next members */
	  UPDATE_CURRENT();

#ifdef FORTRAN_LANG
	  if ( isSgProgHedrStmt (current_stmt) )
	  {
	       down_namescope(SOFT_AND_HARD);

	       namescope->grab_all_symbols (current_stmt);
	  }
#endif

	  process (current_stmt);

          /* Descend the children of the current statement, if it
             has any. */
          
	  /* Child List 1 */

	  if (current_stmt -> numberOfChildrenList1() > 0)
	  {
#          ifndef FORTRAN_LANG
	       down_namescope(SOFT_AND_HARD);
#          endif

	       iterate (current_stmt->childList1(0));

	       /* Restore the current and next members */
	       UPDATE_CURRENT();

#          ifndef FORTRAN_LANG
	       up_namescope(SOFT_AND_HARD);
#          endif
	  }

	  /* Child List 2 */

	  if (current_stmt -> numberOfChildrenList2() > 0)
	  {
#          ifndef FORTRAN_LANG
	       down_namescope(SOFT_AND_HARD);
#          endif

	       iterate (current_stmt->childList2(0));

	       /* Restore the current and next members */
	       UPDATE_CURRENT();

#          ifndef FORTRAN_LANG
	       up_namescope(SOFT_AND_HARD);
#          endif
	  }


	  
#ifdef FORTRAN_LANG
	  if ( isSgProgHedrStmt (current_stmt) )
	  {
	       up_namescope(SOFT_AND_HARD);
	  }
#endif
          /* Then continue along the current childlist */
          
          current_stmt = next_stmt;
     } /* end while (current_stmt != NULL) */
}

