/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


# include <stdio.h>
# include <stdlib.h>
# include <signal.h>

# include "sage++user.h"
# include "sage++callgraph.h"
# include "baseClasses.h"
# include "sage++classhierarchy.h"

# ifndef __GNUC__
#   include <unistd.h>
#else
extern "C" int unlink(const char *path);
extern "C" int access(char *path, int amode);
# endif

# include "compatible.h"
# ifdef SYS5
#   include <string.h>
# else
#   include <strings.h>
# endif

# ifndef R_OK
#   define R_OK 04
# endif

/* -- external functions -------------------------------------------------- */
/* -- from: scanfile.c ---------------------------------------------------- */
extern "C" void openFile (char *path);
extern "C" void closeFile (void);
extern "C" void findFuncOrClass (int line, char **result);
extern "C" void findNextChild (char *name, char **result, int marknewpos);
extern "C" void findNextMember (char *name, int line, char **result);

/* -- from: namecheck.c --------------------------------------------------- */
extern "C" void initCommands (void);
extern "C" int isValidFunc (char *filename, char *classname, char *funcname);

/* -- from: readcomfile.y ------------------------------------------------- */
extern "C" int readComFile (char *filename);

/* -- auxillary functions ------------------------------------------------- */
void printUsage (char* execName)
{
  fprintf (stderr,"Usage: %s\n",execName);
  fprintf (stderr,"  [-help | -h] [-usage | -u] [-debug]\n");
  fprintf (stderr,"  [-verbose | -v] [(-mode | -m) (ascii | vt100 | wyse)]\n");
  fprintf (stderr,"  [(-level | -l) level] [-nolib | -n]\n");
  fprintf (stderr,"  [(-commands | -c) comfile]\n");
  fprintf (stderr,"  depfile\n");
  fprintf (stderr,
     "\n'%s -help' prints description of switches, parameters\n", execName);
  exit (1);
}

void printHelp (char* execName)
{
  printf ("Usage: %s\n", execName);
  printf ("  [-h | -help]  Print this help text\n");
  printf ("  [-u | -usage]  Print usage information (command line switches)\n");
  printf ("  [-v | -verbose]  Turn on verbose mode \n");
  printf ("  [-debug]  Print useful debugging information\n");
  printf ("  [-n | -nolib]  Exclude library functions\n");
  printf ("  [(-m | -mode) (ascii | vt100 | wyse)]  Output mode\n");
  printf ("  [(-l | -level) level]  Maximal level printed\n");
  printf ("  [(-c | -commands) comfile]  Use instrumentation command file\n");
  printf ("  depfile\n");
  exit (1);
}

/* -- global variables ---------------------------------------------------- */
# define M_ASCII 0
# define M_VT100 1
# define M_WYSE  2
# define MAXMODE 3

# ifndef TRUE
#   define FALSE 0
#   define TRUE  1
# endif

char *VERT[MAXMODE]   = { "|", "x", "6" };

char *INIT[MAXMODE]   = { "", "\033(0", "\033H\002" };
char *FINISH[MAXMODE] = { "", "\033(B", "\033H\003" };

char *SO[MAXMODE]  = { "*", "\033[7m", "\b\033G4" };
char *SE[MAXMODE]  = { "*", "\033[0m", "\033G0" };

char *NORMAL[MAXMODE] = { " |__", " tqq", " 4::" };
char *LAST[MAXMODE]   = { " |__", " mqq", " 1::" };

# define NODUMP  0
# define CGDUMP  1
# define TXTDUMP 2
# define ALLDUMP 3
# define CHDUMP  4
# define TXTFIND 99

# define SCANNED 77

int dumpMode = NODUMP;
int verboseMode = FALSE;
int debugMode = FALSE;
int mode = M_ASCII;
int maxlevel = 10;
int done     = FALSE;
int noLib    = FALSE;
int showInst = FALSE;
SgCallGraph CG;
int *finish;

# ifdef __GNUC__
void onintr (int signal)
{
  printf (FINISH[mode]);
  exit (1);
}
# endif

/* -- functions for call graph browsing ----------------------------------- */

static char *classtype[] = { "-", "COLL" };

static char *used;
static int hidden[10];
static maxhidden = 0;

static int isHidden (int tag)
{
  for (int i=0; i<maxhidden; i++) if ( hidden[i] == tag ) return (TRUE);
  return (FALSE);
}

static char *filepath = "";

static int DumpFunc (SgSymbol *funcsymb, int used)
{
  SgSymbol *classsymb;
  SgMemberFuncSymb *memsymb;
  SgStatement *body;
  SgSymbol *f;
  char *classname = "-", *filename = "-", *functype = "Ord";
  int funcline = 0, classtag = -2;
  int isMoE = 0, isColl = 0;
  int numcalls = 0;
  char *ptr;

  if ( isHidden (funcsymb->id()) ) return TRUE;

  if ( memsymb = isSgMemberFuncSymb(funcsymb) ) {
    body      = memsymb->body();
    classsymb = memsymb->className();
    classname = classsymb->identifier();

    if ( strncmp (classname, "brk_", 4) == 0 ) return TRUE;

    classtag  = classsymb->id();
    if ( (isColl = (classsymb->variant() == COLLECTION_NAME)) ) {
      if ( isMoE = memsymb->isMethodOfElement() )
        functype = "MoE";
      else
        functype = "MoC";
    }
    else
      functype = "Mem";
  }
  else
    body = funcsymb->body();

  if ( body ) {
    funcline = body->lineNumber();
    filename = body->fileName();
    /* lastline = body->lastNodeOfStmt()->lineNumber(); */
  }

  if ( dumpMode == TXTFIND ) {
    if ( strcmp (filename, filepath) != 0 ) {
      dumpMode = TXTDUMP;
      filepath = filename;
      if ( *filepath != '-' ) openFile (filepath);
    }
    else
      return FALSE;
  }

  if ( dumpMode != TXTDUMP || strcmp (filename, filepath) == 0 ) {
    if ( dumpMode == TXTDUMP && *filepath != '-' ) {
      findFuncOrClass (funcline, &ptr);
    }
    else
      ptr = "";
     
    printf ("  ftag: %d %s\n", funcsymb->id(), funcsymb->identifier());
    printf (" ffile: %d %s %s\n", funcline, filename, ptr);
    printf ("fcalls: %d -", CG.numberOfCallSites(funcsymb));
    numcalls = 0;
    for (f=CG.firstCalledFunction(funcsymb); f;
         f=CG.nextCalledFunction(funcsymb)) {
      if ( ! isHidden (f->id()) ) {
        printf (" %d", f->id());
        numcalls++;
      }
    }
    printf ("\n");
    if ( dumpMode == TXTDUMP && *filepath != '-' ) {
      printf ("  fpos:");
      for (f=CG.firstCalledFunction(funcsymb); f;
           f=CG.nextCalledFunction(funcsymb)) {
        if ( ! isHidden (f->id()) ) {
          findNextChild (f->identifier(), &ptr, TRUE);
          printf (" { %s }", ptr);
        }
      }
      printf ("\n");
    } else if ( dumpMode == CGDUMP && numcalls ) {
      printf (" fline:");
      for (f=CG.firstCalledFunction(funcsymb); f;
           f=CG.nextCalledFunction(funcsymb)) {
        if ( ! isHidden (f->id()) )
          printf (" %d", CG.SgCalledFunctionStmt(funcsymb)->lineNumber());
      }
      printf ("\n");
    }
    printf ("fclass: %s %s %d\n", classname, classtype[isColl], classtag);
    printf (" ftype: %s %s %s\n", (CG.isCalledInPar(funcsymb) ? "par" : "seq"),
                                 functype, (used ? "used" : "not"));
    printf ("\n");
    return TRUE;
  }
  else
    return FALSE;
}

void TransTree (SgSymbol *func)
{
  int id = CG.functionId (func);
  SgSymbol *f;

  if ( used[id] )
    return;
  else {
    used[id] = TRUE;
    for (f=CG.firstCalledFunction(func); f; f=CG.nextCalledFunction(func))
      TransTree (f);
  }
}

void PrintFunc (SgSymbol *funcsymb)
{
  SgSymbol *classsymb;
  SgMemberFuncSymb *memsymb;
  char *funcname, *classname = 0, *filename = 0;
  char *INS = "", *INE = "";
  int funcline = -1;
  int isMoE = 0, isColl = 0;

  funcname = funcsymb->identifier();
  if ( funcsymb->body() ) {
    funcline = funcsymb->body()->lineNumber();
    filename = funcsymb->body()->fileName();
  }

  if ( memsymb = isSgMemberFuncSymb(funcsymb) ) {
    classsymb = memsymb->className();
    classname = classsymb->identifier();
    if ( (isColl = (classsymb->variant() == COLLECTION_NAME)) )
      isMoE  = memsymb->isMethodOfElement();
  }

  if ( showInst &&
       isValidFunc (filename, classname, funcname) &&
       funcsymb->body() ) {
    INS = SO[mode];
    INE = SE[mode];
  }

  if ( classname ) {
    printf (" %s%s::%s%s", INS, classname, funcname, INE);
    if ( verboseMode )
      printf ("%s%s", (isColl ? " [Coll]" : ""), (isMoE ? " [MoE]" : ""));
  }
  else
    printf (" %s%s%s", INS, funcname, INE);
  if ( verboseMode && filename )
    printf (" [%s:%d]", filename, funcline);
}

void PrintGraph (SgSymbol *func, int level)
{
  int i, j, n, last, no;
  SgSymbol *f;

  if ( no = CG.numberOfFunsCalledFrom(func) ) {
    if (level >= maxlevel) {
      printf (" ...\n");
      return;
    }

    if ( noLib ) {
      n = 0;
      for (f=CG.firstCalledFunction(func), i=0; f;
           f=CG.nextCalledFunction(func), i++) {
        if ( f->body() ) {
          n++;
          last = i;
        }
      }
      if ( n == 0 ) {
        putchar ('\n');
        return;
      }
    }
    else
      last = no - 1;

    finish[level] = FALSE;

    putchar ('\n');
    if (level) {
      printf (INIT[mode]);
      for (j=0; j<=level; j++)
        printf (" %s  ", (finish[j] ? " " : VERT[mode]));
      putchar ('\n');
      printf (FINISH[mode]);
    }

    for (f=CG.firstCalledFunction(func), i=0; f;
         f=CG.nextCalledFunction(func), i++) {
      if ( !noLib || f->body() ) {
        printf (INIT[mode]);
        for (j=0; j<level; j++)
          printf (" %s  ", (finish[j] ? " " : VERT[mode]));
        if (i == last)
          printf (LAST[mode]);
        else
          printf (NORMAL[mode]);
        printf (FINISH[mode]);

        PrintFunc (f);
        done = FALSE;
        if (i == last) finish[level] = TRUE;
        PrintGraph (f, level+1);
      }
    }
    if ( !done ) {
      if (level) {
        printf (INIT[mode]);
        for (j=0; j<level; j++)
          printf (" %s  ", (finish[j] ? " " : VERT[mode]));
        putchar ('\n');
        printf (FINISH[mode]);
      }
      done = TRUE;
    }
  }
  else {
    putchar ('\n');
  }
}

/* -- functions for call graph browsing ----------------------------------- */
static char *type1 = "?iuoe?iuoe";
static char *type2 = "fffffvvvvv";

char *printType (memberType type, SgSymbol *symb, char *cname)
{
  SgDescriptType *d;
  static char res[8];
  char *mname = symb->identifier();

  res[0] = '{';
  res[1] = type1[(int) type];
  res[2] = type2[(int) type];
  res[3] = ' ';
  res[4] = ' ';
  if ( d = isSgDescriptType(symb->type()) ) {
    int attr = d->modifierFlag();
    if ( attr & BIT_VIRTUAL ) res[3] = 'v';
    if ( attr & BIT_INLINE  ) res[4] = 'i';
  }
  if ( strcmp (mname, cname) == 0 )
    res[5] = 'c';
  else if ( mname[0] == '~' && strcmp (mname+1, cname) == 0 )
    res[5] = 'd';
  else if ( (mname[0] < 'a' || mname[0] > 'z') &&
            (mname[0] < 'A' || mname[0] > 'Z') )
    res[5] = 'o';
  else
    res[5] = ' ';
  res[6] = '}';
  res[7] = '\0';

  return res;
}

void TransHierarchy (int level, classHierarchy *t)
{
  classHierarchy *curr;
  classMember *mem;

  if ( level ) {
    if ( strcmp (t->className, "Profiler") == 0 ||
         strncmp (t->className, "brk_", 4) == 0 ) return;

    printf (" ctag: %d %s %s\n", t->classSymbol->id(), t->className,
                                 classtype[t->isCollection()]);
  }
  else
    printf (" ctag: -1 {} -\n");

  printf (" csub:");
  curr = (classHierarchy *) t->children->searchList();
  while ( curr ) {
    if ( strcmp (curr->className, "Profiler") != 0 &&
         strncmp (curr->className, "brk_", 4) != 0 )
      printf (" %d", curr->classSymbol->id());
    curr = (classHierarchy *) t->children->nextItem();
  }
  printf ("\n");

  mem = (classMember *) t->classMembers->searchList();
  while ( mem ) {
    if ( strncmp (mem->name, "brk_", 4) != 0 ) {
      printf (" cmem: %s %d %s\n", mem->name, mem->symbol->id(),
              printType (mem->typeVariant, mem->symbol, t->className));
    }
    mem = (classMember *) t->classMembers->nextItem();
  }
  printf ("\n");

  curr = (classHierarchy *) t->children->searchList();
  while ( curr ) {
    TransHierarchy (level+1, curr);
    curr = (classHierarchy *) t->children->nextItem();
  }
}

void DumpClasses (int level, classHierarchy *t)
{
  classHierarchy *curr;
  classMember *mem;
  char *ptr;

  if ( level ) {
    /* -- hide instr generated classes --------------------------- */
    if ( strcmp (t->className, "Profiler") == 0 ||
         strncmp (t->className, "brk_", 4) == 0 ) return;

    if ( strcmp (t->fileName(), filepath) == 0 ) {
      findFuncOrClass (t->lineNumber(), &ptr);
      printf ("  ctag: %d %s %s\n", t->classSymbol->id(), t->className,
                                    classtype[t->isCollection()]);
      printf (" cfile: %d %s %s\n", t->lineNumber(), t->fileName(), ptr);

      printf (" cbase:");
      curr = (classHierarchy *) t->parents->searchList();
      while ( curr ) {
        if ( strcmp (curr->className, "Profiler") != 0 &&
             strncmp (curr->className, "brk_", 4) != 0 ) {
          findNextMember (curr->className, t->lineNumber(), &ptr);
          printf (" { %d %s }", curr->classSymbol->id(), ptr);
        }
        curr = (classHierarchy *) t->parents->nextItem();
      }
      printf ("\n");

      printf ("  cpos:");
      findNextChild ("public", &ptr, FALSE);          printf (" { %s }", ptr);
      findNextChild ("protected", &ptr, FALSE);       printf (" { %s }", ptr);
      findNextChild ("private", &ptr, FALSE);         printf (" { %s }", ptr);
      findNextChild ("MethodOfElement", &ptr, FALSE); printf (" { %s }", ptr);

      printf ("\n");

      printf (" cmpos:");
      mem = (classMember *) t->classMembers->searchList();
      while ( mem ) {
	if ( type2[mem->typeVariant] == 'f' ) { //is funtion
          if ( strncmp (mem->name, "brk_", 4) != 0 ) {
            findNextMember (mem->name, mem->declaration->lineNumber(), &ptr);
            printf (" { %d %s }", mem->symbol->id(), ptr);
          }
        }
        mem = (classMember *) t->classMembers->nextItem();
      }
      printf ("\n\n");
    }
  }

  curr = (classHierarchy *) t->children->searchList();
  while ( curr ) {
    DumpClasses (level+1, curr);
    curr = (classHierarchy *) t->children->nextItem();
  }
}

/* -- M A I N  function --------------------------------------------------- */

main (int argc, char **argv)
{
  SgProject *project = 0;
  classHierarchy CH;
  char *projFile = NULL;
  char *depFile = NULL;
  char *p;
  FILE *pf = 0;
  int i, l;
  int scanned;
  SgSymbol *mainSymb = 0, *f;

  if (p = getenv("TERM")) {
    if ( (strcmp (p, "vt100") == 0) || (strcmp (p, "ansi") == 0) ||
         (strcmp (p, "xterm") == 0) )
      mode = M_VT100;
    else if ( (p[0] == 'w') && (p[1] == 'y') )
      mode = M_WYSE;
    else
      mode = M_ASCII;
  }

  if (argc == 1) {
    printUsage (argv[0]);
    exit (1);
  }
  else {
    initCommands ();
    for (i=1; i<argc; i++) {
      if ( argv[i][0] == '-' ) {
        p = argv[i];
        if ( (!strcmp(p,"-help")) || (!strcmp(p,"-h")) ) {
          printHelp (argv[0]);
        }
        else if ( (!strcmp(p,"-usage")) || (!strcmp(p,"-u")) ) {
          printUsage (argv[0]);
        }
        else if ( !strcmp(p,"-debug") ) {
          debugMode = TRUE;
        }
        else if ( !strcmp(p,"-dumpall") ) {
          dumpMode = ALLDUMP;
        }
        else if ( !strcmp(p,"-dumptxt") ) {
          dumpMode = TXTDUMP;
        }
        else if ( !strcmp(p,"-dumpcg") ) {
          dumpMode = CGDUMP;
        }
        else if ( !strcmp(p,"-dumpch") ) {
          dumpMode = CHDUMP;
        }
        else if ( (!strcmp(p,"-verbose")) || (!strcmp(p,"-v")) ) {
          verboseMode = TRUE;
        }
        else if ( (!strcmp(p,"-nolib")) || (!strcmp(p,"-n")) ) {
          noLib = TRUE;
        }
        else if ( (!strcmp(p,"-mode")) || (!strcmp(p,"-m")) ) {
          if ( i == (argc-1) ) {
            fprintf (stderr,"ERROR: not enough arguments for %s switch\n", p);
            printUsage (argv[0]);
          }
          i++;
          switch ( argv[i][0] ) {
            case 'a':
            case 'A': mode = M_ASCII;
                      break;
            case 'v':
            case 'V': mode = M_VT100;
                      break;
            case 'w':
            case 'W': mode = M_WYSE;
                      break;
            default : /*ignore*/
                      break;
          }
        }
        else if ( (!strcmp(p,"-level")) || (!strcmp(p,"-l")) ) {
          if ( i == (argc-1) ) {
            fprintf (stderr,"ERROR: not enough arguments for %s switch\n", p);
            printUsage (argv[0]);
          }
          i++;
          maxlevel = atoi(argv[i]);
        }
        else if ( (!strcmp(p,"-commands")) || (!strcmp(p,"-c")) )
        {
          if ( i == (argc-1) )
          {
            fprintf (stderr,"ERROR: not enough arguments for %s switch\n", p);
           printUsage (argv[0]);
          }
          i++;
          if ( verboseMode ) printf ("reading commands file %s\n", argv[i]);
          if ( ! readComFile (argv[i]) ) exit (1);
          showInst = TRUE;
        }
        else {
          fprintf (stderr,"ERROR: unknown option `%s'\n", p);
          printUsage (argv[0]);
        }
      }
      else {
        depFile = argv[i];
        l = strlen(argv[i]);
        p = argv[i] + l - 4;
        if ( strcmp (p, ".dep") != 0 ) {
          fprintf (stderr,"ERROR: `%s': no dep file\n", argv[i]);
          exit (1);
        }
        else {
          *p = '\0';
          projFile = new char[l+2];
          sprintf (projFile, "%s.proj", argv[i]);
          *p = '.';
        }

        if ( access (projFile, R_OK) == -1 ) {
          pf = fopen (projFile, "w");
          fprintf (pf, "%s\n", depFile);
          fclose (pf);
        }

        if ( verboseMode || debugMode )
          printf ("using project file %s\n", projFile);
        project = new SgProject (projFile);
      }
    }
  }

  if ( depFile ) {
    if ( verboseMode || debugMode )
      printf ("processing %s:\n", project->fileName(0));

    if ( dumpMode ) {
      if ( dumpMode == CHDUMP ) {
        CH.determineClassHierarchy(&(project->file(0)));
        CH.sort();
        TransHierarchy (0, &CH);

      } else {
        CG.GenCallTree(&(project->file(0)));
        CG.computeClosures();

        int num = CG.numberOfFunctionsInGraph();
        printf ("%d\n\n", num);

        used = new char[num];
        for (f=CG.firstFunction(), i=0; f; f=CG.nextFunction(), i++) {
          used[i] = FALSE;
          p = f->identifier();
          if ( strcmp (p, "Processor_Main") == 0 || strcmp (p, "main") == 0 ) {
            mainSymb = f;
          } else if ( strcmp (p, "Profiler") == 0 ||
                      strcmp (p, "~Profiler") == 0 ||
                      strcmp (p, "PROFILE_COLL_NAMES") == 0 ||
                      strncmp (p, "brk_", 4) == 0 ) {
            hidden[maxhidden++] = f->id();
          }
        }
        if (!mainSymb) {
          fprintf (stderr, "%s: no main function\n", depFile);
          exit (1);
        }
        TransTree (mainSymb);

        if ( dumpMode == CGDUMP ) {
          for (f=CG.firstFunction(), i=0; f; f=CG.nextFunction(), i++) {
            if ( used[i] ) DumpFunc (f, TRUE);
          }
        }
        else if ( dumpMode == ALLDUMP ) {
          for (f=CG.firstFunction(), i=0; f; f=CG.nextFunction(), i++) {
            DumpFunc (f, used[i]);
          }
        }
        else if ( dumpMode == TXTDUMP ) {
          CH.determineClassHierarchy(&(project->file(0)));

          scanned = 0;
          while ( scanned != num ) {
            dumpMode = TXTFIND;
            for (f=CG.firstFunction(), i=0; f; f=CG.nextFunction(), i++) {
              if ( used[i] != SCANNED ) {
                if ( DumpFunc (f, used[i]) ) {
                  used[i] = SCANNED;
                  scanned++;
                }
              }
            }
            if ( *filepath != '-' ) DumpClasses (0, &CH);
          }
          closeFile ();
        }
        delete [] used;
      }
    }
    else {
      CG.GenCallTree(&(project->file(0)));
      CG.computeClosures();

      for (f=CG.firstFunction(); f; f=CG.nextFunction()) {
        if ( strcmp (f->identifier(), "Processor_Main") == 0 ) {
          mainSymb = f;
          break;
        }
      }

# ifdef __GNUC__
      signal (SIGINT,  onintr);
      signal (SIGQUIT, onintr);
# endif

      finish = new int[maxlevel+1];
      for (i=0; i<=maxlevel; i++) finish[i] = TRUE;

      putchar ('\n');
      PrintFunc (mainSymb);
      printf ("\n %s%s%s", INIT[mode], VERT[mode], FINISH[mode]);
      PrintGraph (mainSymb, 0);
      putchar ('\n');
    }

    if ( pf ) unlink(projFile);
  }
  exit (0);
  return 0; // to make the compiler happy
}
